------------------------------------------------------------------------------
--                                                                          --
--                GNU ADA RUN-TIME LIBRARY (GNARL) COMPONENTS               --
--                                                                          --
--                  S Y S T E M . O S _ P R I M I T I V E S                 --
--                                                                          --
--                                  B o d y                                 --
--                                                                          --
--                             $Revision: 1.9 $                             --
--                                                                          --
--          Copyright (C) 1998-1999 Free Software Foundation, Inc.          --
--                                                                          --
-- GNARL is free software; you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion. GNARL is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNARL; see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
-- GNARL was developed by the GNARL team at Florida State University. It is --
-- now maintained by Ada Core Technologies Inc. in cooperation with Florida --
-- State University (http://www.gnat.com).                                  --
--                                                                          --
------------------------------------------------------------------------------

--  This is the NT version of this package

with System.OS_Interface;

package body System.OS_Primitives is

   use System.OS_Interface;

   ---------------------------------------
   -- Data for the high resolution clock --
   ---------------------------------------

   Tick_Frequency      : aliased LARGE_INTEGER;
   --  Holds frequency of high-performance counter used by Clock
   --  Windows NT uses a 1_193_182 Hz counter on PCs.

   Base_Ticks          : aliased LARGE_INTEGER;
   --  Holds the Tick count for the base time.

   Base_Clock          : Duration;
   --  Holds the current clock for the base time.

   procedure Get_Base_Time;
   --  Retreive the base time. This base time will be used by clock to
   --  compute the current time by adding to it a fraction of the
   --  performance counter. This is for the implementation of a
   --  high-resolution clock.

   -----------
   -- Clock --
   -----------

   --  this is a high resolution clock implemented using
   --  QueryPerformanceCounter. This call return a 64 bits values (based on
   --  the 8253 16 bits counter). This counter is updated every 1/1_193_182
   --  times per seconds. The call to QueryPerformanceCounter takes 6
   --  microsecs to complete.

------------------------------------------------------------------------------
--                                                                          --
-- Ada.Calendar.Clock and Ada.Real_Time.Clock may leap forward by several   --
-- seconds.  This patch attempts to correct such behavior.  See             --
-- http://support.microsoft.com/directory/article.asp?ID=KB;EN-US;Q274323&  --
--                                                                          --
-- A simple test for the problem on a particular system is to time:         --
--   End_Time := Ada.Calendar.Clock + 10.0;                                 --
--   while Ada.Calendar.Clock < End_Time loop null;end loop;                --
-- and see whether it takes 10 seconds, or a fraction of a second, to run.  --
-- Tom Moran    tmoran@acm.org                                              --
--                                                                          --
------------------------------------------------------------------------------
   package QPC_Fix is
      function Query_Performance_Counter (P : access LARGE_INTEGER)
      return BOOL;
   end QPC_Fix;

   function Clock return Duration is
      Current_Ticks : aliased LARGE_INTEGER;
   begin
--    if not QueryPerformanceCounter (Current_Ticks'Access) then
      if not QPC_Fix.Query_Performance_Counter (Current_Ticks'Access) then

         return 0.0;
      end if;

      return Base_Clock +
        Duration (Long_Long_Float (Current_Ticks - Base_Ticks) /
                  Long_Long_Float (Tick_Frequency));
   end Clock;

   -------------------
   -- Get_Base_Time --
   -------------------

   procedure Get_Base_Time is
      use System.OS_Interface;

      --  the resolution for GetSystemTime is 1 millisec.
      --  the time to get both base times should take less that 1 millisec.
      --  Therefore, the elapsed time reported by GetSystemTime between both
      --  actions should be null.

      Max_Elapsed    : constant := 0;

      Systime        : aliased SYSTEMTIME;
      Test_Systime   : aliased SYSTEMTIME;

      Now            : aliased Long_Long_Integer;
      Test_Now       : aliased Long_Long_Integer;

      epoch_1970     : constant := 16#19D_B1DE_D53E_8000#;  -- win32 UTC epoch
      system_time_ns : constant := 100;   -- 100 ns per tick
      Sec_Unit       : constant := 10#1#E9;

   begin
      --  Here we must be sure that both of these calls are done in a short
      --  amount of time. Both are base time and should in theory be taken
      --  at the very same time.

      loop
         GetSystemTime (Systime'Access);

--       if not QueryPerformanceCounter (Base_Ticks'Access) then
         if not QPC_Fix.Query_Performance_Counter (Base_Ticks'Access) then
            pragma Assert
              (Standard.False,
               "Could not query high performance counter in Clock");
            null;
         end if;

         GetSystemTime (Test_Systime'Access);

         --  compute the time between both call to GetSystemTime and try again
         --  if the time between both are above Max_Elapsed.

         if not SystemTimeToFileTime (Systime'Access, Now'Access) then
            pragma Assert
              (Standard.False,
               "Call to SystemTimeToFileTime failed");
            null;
         end if;

         if SystemTimeToFileTime (Test_Systime'Access, Test_Now'Access) then
            pragma Assert
              (Standard.False,
               "Call to SystemTimeToFileTime failed");
            null;
         end if;

         exit when (Test_Now - Now) = Max_Elapsed;
      end loop;

      Base_Clock := Duration
        (Long_Long_Float ((Now - epoch_1970) * system_time_ns) /
         Long_Long_Float (Sec_Unit));
   end Get_Base_Time;

   -----------------
   -- Timed_Delay --
   -----------------

   procedure Timed_Delay (Time : Duration; Mode : Integer) is
      Rel_Time   : Duration;
      Abs_Time   : Duration;
      Check_Time : Duration := Clock;

   begin
      if Mode = Relative then
         Rel_Time := Time;
         Abs_Time := Time + Check_Time;
      else
         Rel_Time := Time - Check_Time;
         Abs_Time := Time;
      end if;

      if Rel_Time > 0.0 then
         loop
            Sleep (DWORD (Rel_Time * 1000.0));
            Check_Time := Clock;

            exit when Abs_Time <= Check_Time;

            Rel_Time := Abs_Time - Check_Time;
         end loop;
      end if;
   end Timed_Delay;

   package body QPC_Fix is

      Prev_Returned_PC : LARGE_INTEGER := 0.0;
      --  return 0.0 if initial Query doesn't work

      PC_Correction,
      Prev_PC : LARGE_INTEGER := 0.0;

      Red_Flag : LARGE_INTEGER := LARGE_INTEGER'Last;
      --  set at first use to 1 second

      Flag_Value : constant LARGE_INTEGER := 16777216.0; -- 2**24, fixed pt

      type Approx_Float is digits 15;
      --  The accuracy we need is only for matching to GetTickCount,
      --  which has a 10ms granularity, so roughly the righmost 15
      --  bits of a performance_count are unimportant.  Thus we need
      --  at most 64-15 = 49 bits which fits happily within the 52 bit
      --  mantissa of the 8087 8-byte float format.  At a frequency of
      --  3MHz, a year only counts to 2**47, so we need only 47-15 = 32
      --  except for very long running systems with much faster clocks.

      Freq : aliased Approx_Float;
      Flag_ms : Approx_Float;

      Base_PC : aliased LARGE_INTEGER;
      Base_Tick : DWORD;

      function GetTickCount return DWORD;
      pragma Import (Stdcall, GetTickCount, "GetTickCount");

      Need_Initialization : Boolean := True;

      procedure Initialize;

      procedure Initialize is
         F : aliased LARGE_INTEGER;
      begin
         Need_Initialization := False;
         --  If the system returns 0 for the clock functions here,
         --  Red_Flag will remain >= any clock change, and QPC_Fix
         --  will effectively do nothing.
         if QueryPerformanceCounter (Base_PC'Access)
         and then QueryPerformanceFrequency (F'Unchecked_Access)
         and then F /= 0.0 then
            Base_Tick := GetTickCount;
            PC_Correction := 0.0;
            Prev_PC := Base_PC;
            Prev_Returned_PC := Prev_PC;
            Red_Flag := F; -- let's use one second as a cutoff value
            Freq := Approx_Float (F);
            Flag_ms := Approx_Float (Flag_Value) * 1000.0 / Freq;
         end if;
      end Initialize;

      function Query_Performance_Counter (P : access LARGE_INTEGER)
      return BOOL is
         Now_Tick : DWORD;
         Now_PC  : aliased LARGE_INTEGER;
         Corrected_PC, Change : LARGE_INTEGER;
         Error_ms, Change_ms, Tick_ms : Approx_Float;  -- all in milliseconds
         Returnee : BOOL;
      begin
         if Need_Initialization then
            Initialize;
         end if;
         Returnee := QueryPerformanceCounter (Now_PC'Unchecked_Access);
         if not Returnee then
            P.all := Prev_Returned_PC;
            return Returnee;
         end if;
         Change := Now_PC - Prev_PC;
         Prev_PC := Now_PC;
         if Change >= Red_Flag then -- big change, raise red flag
            Change_ms := Approx_Float (Now_PC - Base_PC) * 1000.0 / Freq;
            Now_Tick := GetTickCount;
            Tick_ms := Approx_Float (Now_Tick - Base_Tick);
            -- Note: Now_Tick - Base_Tick is a mod 2**32 value and
            -- will have rolled over if it's been over 49.7 days
            -- since Base_Tick.
            -- If two successive calls are more than 49.7 days apart,
            -- this will be erroneous.  If that's a possibility with
            -- your app, you'll have to check with the mmddyy clock
            -- and correct appropriately.  That situation seems unlikely.
            Error_ms := Change_ms - Tick_ms;
            while Error_ms > 2.0**32 loop   -- adjust for apparent rollover
              Error_ms := Error_ms - 2.0**32;
            end loop;
            --  Note: Error_ms precision same as Tick_ms, 10ms
            --  Check/adjust for (near) Flag_Value jump
            while Error_ms > Flag_ms - 10.0 loop
               PC_Correction := PC_Correction + Flag_Value;
               Error_ms := Error_ms - Flag_ms;
            end loop;
            if Error_ms not in -500.0 .. 500.0 then
               --  Still a big difference between clocks, not explained by
               --  Flag_Value size leap.  We'll assume GetTickCount correct,
               --  and adjust PC_Correction.
               PC_Correction := PC_Correction
                                + LARGE_INTEGER (Error_ms * Freq / 1000.0);
            end if;
            Base_PC := Now_PC;
            Base_Tick := Now_Tick;
         end if;
         Corrected_PC := Now_PC - PC_Correction;
         if Corrected_PC >= Prev_Returned_PC then -- clock must not backup
            Prev_Returned_PC := Corrected_PC;
         end if;
         P.all := Prev_Returned_PC;
         return Returnee;
      end Query_Performance_Counter;

   end QPC_Fix;

begin
   if not QueryPerformanceFrequency (Tick_Frequency'Access) then
      pragma Assert
        (Standard.False,
         "cannot get high performance counter frequency");
      null;
   end if;

   Get_Base_Time;

end System.OS_Primitives;
